"""Creates the built-in 'smtp' connector that can be used to forward SMTP message to an upstream SMTP server."""

import smtplib
import socket

from mail2beyond import framework


class Connector(framework.BaseConnector):
    """Defines a connector that forwards received messages to an upstream SMTP server for actual delivery."""
    name = "smtp"

    def submit(self, parser):
        """Overwrites the submit() method to forward the received SMTP message to an upstream SMTP server."""
        # Setup the SMTP client with or without TLS, depending on the configuration
        if self.config.get("smtp_use_tls", False):
            client = smtplib.SMTP_SSL(self.config.get("smtp_host"), self.config.get("smtp_port"))
        else:
            client = smtplib.SMTP(self.config.get("smtp_host"), self.config.get("smtp_port"))

        # Add login credentials to the client if configured
        if self.config.get("smtp_use_login", False):
            client.login(self.config["smtp_login_user"], self.config["smtp_login_password"])

        # Gather information about SMTP connection
        proto = "smtps" if self.config.get("smtp_use_tls", False) else "smtp"
        host = self.config.get("smtp_host")
        port = self.config.get("smtp_port")

        # Send the SMTP message
        try:
            client.sendmail(
                parser.mail.envelope.mail_from,
                parser.mail.envelope.rcpt_tos,
                parser.mail.envelope.content.decode()
            )
            client.close()
            self.log.debug(f"connector '{self}' successfully forwarded message to {proto}://{host}:{port}")
        except Exception as sendmail_error:
            self.log.error(f"connector '{self}' failed to forward message - {sendmail_error}")
            client.close()
            raise sendmail_error

    def pre_submit(self, parser):
        """Overwrites the pre_submit() method to ensure required configuration is set."""
        # Require an 'smtp_host' value to be specified
        if "smtp_host" not in self.config:
            raise framework.Error(f"connector '{self}' requires config value 'smtp_host'")

        # Ensure 'smtp_host' is a str
        if not isinstance(self.config.get("smtp_host"), str):
            raise framework.Error(f"connector '{self}' config value 'smtp_host' must be type 'str'")

        # Require 'smtp_host' to be a valid IP address or hostname
        try:
            socket.gethostbyname(self.config.get("smtp_host"))
        except socket.gaierror as err:
            raise framework.Error(f"connector '{self}' contains invalid 'smtp_host' : {err}")

        # Require an 'smtp_port' value to be specified
        if "smtp_port" not in self.config:
            raise framework.Error(f"connector '{self}' requires config value 'smtp_port'")

        # Require a 'smtp_port' to be integer between 1 adn 65535
        if not isinstance(self.config.get("smtp_port"), int) or 1 > self.config.get("smtp_port") > 65535:
            raise framework.Error(f"connector '{self}' requires config value 'smtp_port' to be between 1 and 65535")

        # Only check the 'smtp_use_tls' value if it is specified otherwise the default will be assumed.
        if "smtp_use_tls" in self.config:
            # Ensure 'smtp_use_tls' is a bool
            if not isinstance(self.config.get("smtp_use_tls"), bool):
                raise framework.Error(f"connector '{self}' config value 'smtp_use_tls' must be type 'bool'")

        # Only check the 'smtp_use_login' requirements if it is specified, otherwise no login will be applied
        if self.config.get("smtp_use_login", False):
            # Require an 'smtp_login_user' value to be specified
            if "smtp_login_user" not in self.config:
                raise framework.Error(f"connector '{self}' requires config value 'smtp_login_user'")

            # Ensure 'smtp_login_user' is a str
            if not isinstance(self.config.get("smtp_login_user"), str):
                raise framework.Error(f"connector '{self}' config value 'smtp_login_user' must be type 'str'")

            # Require an 'smtp_login_password' value to be specified
            if "smtp_login_password" not in self.config:
                raise framework.Error(f"connector '{self}' requires config value 'smtp_login_password'")

            # Ensure 'smtp_login_password' is a str
            if not isinstance(self.config.get("smtp_login_password"), str):
                raise framework.Error(f"connector '{self}' config value 'smtp_login_password' must be type 'str'")



"""
Module that contains a parser that automatically selects another parser based on the mail's content-type header.
"""

from mail2beyond import framework
from . import plain, html


class Parser(framework.BaseParser):
    """Creates a parser to select the parser that matches this mail's content-type header."""
    name = "auto"

    def __init__(self, mail):
        """Initialize this Parser object."""
        super().__init__(mail)

    def parse_content(self):
        """Automatically selects the parser to use based on the mail's content-type header and parses the content."""
        return self.get_parser_by_content_type().content

    def get_parser_by_content_type(self):
        """
        Uses the mail's 'content-type' header to determine which parser to use. In the case there is no match, the
        default 'plain' parser will be used.

        Returns:
            mail2beyond.framework.BaseParser: The Parser object determined to be the best match for the content-type.
        """
        # Fetch our content type
        content_type = self.mail.headers.get_content_type()

        # Use the html parser if html content type
        if self.mail.headers.get_content_type().lower() == "text/html":
            self.log.debug(f"auto parser found 'html' match for content-type '{content_type}'")
            return html.Parser(self.mail)
        # Use the plain parser if plain content type
        if self.mail.headers.get_content_type().lower() == "text/plain":
            self.log.debug(f"auto parser found 'plain' match for content-type '{content_type}'")
            return plain.Parser(self.mail)

        # Assume the text parser if no match was found
        self.log.debug(f"auto parser found no match for content-type '{content_type}' using default 'plain' parser")
        return plain.Parser(self.mail)



"""
Module that contains a parser that parses an email's content body as HTML and converts it to a markdown format.
"""

import html2text

from mail2beyond import framework


class Parser(framework.BaseParser):
    """Creates a parser to parse an email's content body as HTML."""
    name = "html"

    def __init__(self, mail):
        """Initialize this Parser object."""
        super().__init__(mail)

    def parse_content(self):
        """Converts HTML content body to markdown formatted string."""
        return html2text.html2text(html=self.mail.content, bodywidth=len(self.mail.content))



"""
.. include:: ../docs/DOCUMENTATION.md
"""
from . import framework
from . import connectors
from . import parsers
from . import tools

# Set the version of this package
__version__ = "1.0.2"

# Don't include tests module in generated documentation.
__pdoc__ = {"tests": False}



"""Module that groups built-in connectors for mail2beyond to use."""
from . import void
from . import smtp
from . import slack
from . import discord
from . import google_chat
from . import microsoft_teams



"""Module that contains the core framework for mail2beyond."""

import email
import inspect
import ipaddress
import re
import logging
import signal
import ssl

from aiosmtpd.controller import Controller


class Error(BaseException):
    """Creates the `Error` object used by mail2beyond."""
    def __init__(self, message: str):
        super().__init__(message)


class Email:
    """
    Creates an `Email` object that contains a decoded SMTP email along with information about the client and server.

    Attributes:
        server (aiosmtpd.smtp.SMTP): The `SMTP` object that handled the email from `aiosmtpd`.
        session (aiosmtpd.smtp.Session): The `Session` object that contains client-connection info from `aiosmtpd`.
        envelope (aiosmtpd.smtp.Envelope): The `Envelope` object that contains the original email as it was received by
            the server from `aiosmtpd`.
        headers (email.message.Message): The Message object from the 'email' Python module that contains the decoded
            SMTP headers.


    """

    def __init__(self, server, session, envelope):
        """
        Initializes the `Email` object with required attributes using parameters.

        Args:
            server (aiosmtpd.smtp.SMTP): The `SMTP` object that handled the email from `aiosmtpd`.
            session (aiosmtpd.smtp.Session): The `Session` object that contains client-connection info from `aiosmtpd`.
            envelope (aiosmtpd.smtp.Envelope): The `Envelope` object that contains the original email as it was received
                by the server from `aiosmtpd`.
        """
        self.server = server
        self.session = session
        self.envelope = envelope
        self.headers = email.message_from_bytes(envelope.content)

    def get_peer_ip(self):
        """Gets the IP of the remote peer (client)."""
        return self.session.peer

    def get_peer_ip_and_port(self):
        """Gets the IP and port of the remote peer (client) in IP:PORT format."""
        return f"{self.session.peer[0]}:{self.session.peer[1]}"

    def get_server_ip_and_port(self):
        """Gets the IP and port of the server that accepted the email in IP:PORT format."""
        return f"{self.server.event_handler.address}:{self.server.event_handler.port}"

    # Getters and setters
    @property
    def content(self):
        """Gets the decoded content body from the email."""
        return self.headers.get_payload(decode=True).decode()


class Listener:
    """
    Creates the `Listener` object that accepts SMTP/SMTPS requests and applies logic based on a specified mappings.

    Attributes:
        log (logging.Logger): The Logger object the `Listener` will use to log events that occur while listening.
        controller (aiosmtpd.controller.Controller): The `Controller` from `aiosmtpd` that controls the SMTP server.
    """
    # Private attributes are not for public consumption.
    # pylint: disable=too-many-instance-attributes

    # Initialize attributes.
    _address = None
    _port = None
    _mappings = None
    _tls_context = None
    _enable_starttls = None
    _require_starttls = None

    def __init__(self, mappings: list, address: str = "127.0.0.1", port: int = 62125, **kwargs):
        """
        Initializes the `Listener` with the required attributes from arguments.

        Args:
            mappings (list): A list of `mail2beyond.framework.Mapping` objects that this listener check whenever an
                SMTP message is received. At list one item is required and one item must be a Mapping with its `pattern`
                attribute set to `default`.
            address (str): The local address the `Listener` will listen for SMTP requests on.
            port (int): The local TCP port the `Listener` will listen for SMTP requests on.
            **tls_context (ssl.SSLContext): The SSLContext object from the 'ssl' library that defines the TLS
                configuration for an SMTPS listener. If `None` is specified, the `Listener` will not use TLS.
            **enable_starttls (bool): Enables or disables allowing advertisement of the STARTTLS option that allows
                clients to automatically upgrade insecure SMTP connections to secure connections. This argument is only
                applicable if you have provided a valid `tls_context`.
            **require_starttls (bool): Enables or disables requiring clients to use the STARTTLS option. If enabled, all
                clients must choose the STARTTLS option or the request will be rejected. This argument is only
                applicable if you have set `enable_starttls` to `True`.
        """
        # Setup logging
        self.log = None
        self.setup_logging(kwargs.get("log_level", logging.NOTSET))

        # Setup mappings and the SMTP controller
        self.controller = None
        self.mappings = mappings
        self.address = address
        self.port = port
        self.tls_context = kwargs.get("tls_context", None)
        self.enable_starttls = kwargs.get("enable_starttls", False)
        self.require_starttls = kwargs.get("require_starttls", False)
        self.setup_controller()

    def start(self):
        """
        Starts the `Listener`. This will run the SMTP server in the background. After calling this method, you will
            need to keep the process alive for the SMTP server to continue receiving SMTP connections. You can call the
        `wait()` method to keep the server running indefinitely, or use a function like time.sleep() to keep the
            SMTP server running for a specified amount of time.
        """
        self.controller.start()
        self.log.info(f"mail2beyond started listening on {self.address}:{self.port}")
        self.log_mappings()

    @staticmethod
    def wait():
        """
        Allows the `Listener` to wait indefinitely, so it can continually accept incoming SMTP messages. This is
        typically called immediately after the `start()` method.
        """
        signal.pause()

    def setup_controller(self):
        """
        Sets up the `Controller` object with the current address, port and options. If you have made changes to your
        `Listener` object, you will need to call this method to reconfigure the `Controller`.

        Raises:
            mail2beyond.framework.Error: When an unexpected error occurs when creating controllers.
        """
        # Setup SMTPS controller if enabled
        if self.tls_context and not self.enable_starttls:
            self.controller = Controller(self, hostname=self.address, port=self.port, ssl_context=self.tls_context)
        # Setup STARTTLS controller if enabled
        elif self.tls_context and self.enable_starttls:
            self.controller = Controller(
                self,
                hostname=self.address,
                port=self.port,
                tls_context=self.tls_context,
                require_starttls=self.require_starttls
            )
        # Setup SMTP controller if no tls_context was provided
        elif not self.tls_context:
            self.controller = Controller(self, hostname=self.address, port=self.port)
        # Throw an error if we somehow ended up here
        else:
            raise Error("an unexpected error occurred creating the Listener controller")

    def setup_logging(self, level: int = logging.NOTSET, handler=None, **kwargs):
        """
        Sets up the `log` attribute with a configurable Logger for this `Listener`.

        Args:
            level (int): Sets the logging level the Logger will start logging at. See
                https://docs.python.org/3/library/logging.html#logging-levels
            handler (logging.Handler): Sets the logging handler to use. You can pass in a custom Handler like a
                logging.FileHandler to log to a file. If no handler is specified, the default handler
                logging.StreamHandler is assumed which will only print logs to the console.
            **log_format (str): Sets the format of log messages. See
                https://docs.python.org/3/library/logging.html#logging.Formatter.format
            **log_date_format (str): Sets the format of datetime strings in log messages. See
                https://docs.python.org/3/library/logging.html#logging.Formatter.formatTime
        """

        # Reset the existing logger
        del self.log

        # Set formatting
        log_format = kwargs.get("log_format", "[%(asctime)s][%(levelname)s]:%(message)s")
        log_date_format = kwargs.get("log_date_format", "%b %d %Y %H:%M:%S")

        # Set handler
        handler = handler if handler else logging.StreamHandler()
        handler.setLevel(level)
        handler.setFormatter(logging.Formatter(fmt=log_format, datefmt=log_date_format))

        # Set the logger
        self.log = logging.getLogger(__name__)
        self.log.setLevel(level)
        self.log.addHandler(handler)

        # Re-assign assigned mappings to re-populate connector loggers
        if self.mappings:
            self.mappings = self.mappings

        # Log debug notice if debug level set
        if level == logging.DEBUG:
            self.log.warning("logging at level DEBUG may expose sensitive information in logs")

    def log_mappings(self):
        """Logs the configured mappings upon startup for debugging."""
        # Loop through all configured mappings and log it's configuration.
        for mapping in self.mappings:
            self.log.debug(
                f"loaded mapping with pattern='{ mapping.pattern }' field='{ mapping.field }'"
                f" connector='{ mapping.connector }' connector_module='{ mapping.connector.__module__ }'"
                f" parser={ mapping.parser }"
            )

    def get_default_mapping(self, mappings: (list, None) = None):
        """
        Gets the mapping with the `default` pattern from a list of mappings.

        Args:
            mappings (list, None): The list of Mapping objects to search. If none are provided, the mappings defined
                in the `mappings` attribute value will be assumed.

        Raises:
            mail2beyond.framework.Error: When no Mapping object within the list has its `pattern` set to `default`.
        """
        # Variables
        mappings = mappings if mappings else self.mappings

        # Loop through each mapping and find the default.
        for mapping in mappings:
            if mapping.pattern == "default":
                return mapping

        # Return nothing if no default was found.
        raise Error("mapping with 'default' pattern is required")

    def get_mapping_matches(self, mail: Email):
        """
        Gets the mappings that match a specific `Email` object.

        Args:
            mail (mail2beyond.framework.Email): The `Email` object to check for Mapping matches.

        Returns
            list: a list of Mapping objects that matched this `Email`.
        """
        # Initialize the default mapping in case there were no direct matches.
        default_mapping = self.get_default_mapping()
        matched_mappings = []

        # Loop through each mapping to check if the condition matches
        for mapping in self.mappings:
            # Only match default as last resort
            if mapping != default_mapping:
                if mapping.is_match(mail.headers.get(mapping.field, None)):
                    matched_mappings.append(mapping)

        # Return the matched mappings if there were any, otherwise return the default mapping.
        if matched_mappings:
            return matched_mappings

        # Otherwise return default mapping
        return [default_mapping]

    async def handle_DATA(self, server, session, envelope):
        """
        Overwrites the `handle_DATA()` method from `aiosmtpd` that defines how received SMTP data is handled. There
        shouldn't be any need for this method to be called explicitly as it is used exclusively by `aiosmtpd`.

        See Also:
            https://aiosmtpd.readthedocs.io/en/latest/handlers.html#handle_DATA

        Args:
            server (aiosmtpd.smtp.SMTP): The `SMTP` object that handled the email from `aiosmtpd`.
            session (aiosmtpd.smtp.Session): The `Session` object that contains client-connection info from `aiosmtpd`.
            envelope (aiosmtpd.smtp.Envelope): The `Envelope` object that contains the original email as it was received
                by the server from `aiosmtpd`.
        """
        # Create an email object with the received session and data
        mail = Email(server, session, envelope)

        # Log this connection
        self.log.info(f"server {mail.get_server_ip_and_port()} connection from peer {mail.get_peer_ip_and_port()}")

        # Pull the mapping that matches this message and send the chat accordingly.
        mappings = self.get_mapping_matches(mail)

        # Loop through each matched mapping and run it's connector
        for mapping in mappings:
            self.log.debug(
                f"running connector '{mapping.connector}' because {mapping.field.upper()} "
                f"'{mail.headers.get(mapping.field)}' matched mapping '{mapping.pattern}'"
            )

            # Run the connector with this mapping's parser
            self.log.debug(f"using parser '{mapping.parser}'")
            mapping.connector.run(mapping.parser(mail))

        return '250 Message accepted'

    # Getters and setters
    @property
    def address(self):
        """The property to get and/or set the  address attribute."""
        return self._address

    @address.setter
    def address(self, value: str):
        """Sets the address attribute after validating the new value."""
        # Require address to be valid IP, or localhost
        if value not in ["localhost"]:
            try:
                ipaddress.ip_address(value)
            except ValueError as exc:
                raise Error("'address' must be valid IP or localhost") from exc

        self._address = value

    @property
    def port(self):
        """The property to get and/or set the  port attribute."""
        return self._port

    @port.setter
    def port(self, value: int):
        """Sets the port attribute after validating the new value."""
        # Require port to be valid TCP port
        if isinstance(value, int) and 1 <= value <= 65535:
            self._port = value
        else:
            raise Error("'port' must be valid TCP port")

    @property
    def mappings(self):
        """The property to get and/or set the  mappings attribute."""
        return self._mappings

    @mappings.setter
    def mappings(self, value: list):
        """Sets the mappings attribute after validating the new value."""
        # Require mappings to be list
        if not isinstance(value, list):
            raise TypeError("'mappings' must be type 'list'")

        # Loop through each requested mapping and it is correct object type and a default object is present
        found_default = False
        for mapping in value:
            # Require mapping to be Mapping object
            if not isinstance(mapping, Mapping):
                raise Error("each 'mappings' item must be 'Mapping' object")

            # Check if this mapping is the default.
            if mapping.pattern == "default":
                # Ensure we have not already found a default
                if not found_default:
                    found_default = True
                else:
                    raise Error("multiple 'mappings' items assigned pattern 'default'")

            # Assign this listener's logger to each mapping connector
            mapping.connector.log = self.log

        # Ensure we found a default mapping
        if found_default:
            self._mappings = value

        else:
            raise Error("missing 'mappings' item with pattern 'default'")

    @property
    def tls_context(self):
        """The property to get and/or set the  tls_context attribute."""
        return self._tls_context

    @tls_context.setter
    def tls_context(self, value: ssl.SSLContext):
        """Sets the tls_context attribute after validating the new value."""
        # Require tls_context to be SSLContext object or None
        if isinstance(value, ssl.SSLContext) or value is None:
            self._tls_context = value
        else:
            raise Error("'tls_context' must be type 'ssl.SSLContext' or 'None'")

    @property
    def enable_starttls(self):
        """The property to get and/or set the  enable_starttls attribute."""
        return self._enable_starttls

    @enable_starttls.setter
    def enable_starttls(self, value: bool):
        """Sets the enable_starttls attribute after validating the new value."""
        # Require enable_starttls to be bool
        if isinstance(value, bool):
            self._enable_starttls = value
        else:
            raise Error("'enable_starttls' must be type 'bool'")

    @property
    def require_starttls(self):
        """The property to get and/or set the  require_starttls attribute."""
        return self._enable_starttls

    @require_starttls.setter
    def require_starttls(self, value: bool):
        """Sets the require_starttls attribute after validating the new value."""
        # Require require_starttls to be bool
        if not isinstance(value, bool):
            raise Error("'require_starttls' must be type 'bool'")

        # Require 'enable_tls' to be True before allowing
        if not self.enable_starttls and value:
            raise Error("'require_starttls' cannot be 'True' while 'enable_starttls' is 'False'")

        self._require_starttls = value


class Mapping:
    """
    Creates a Mapping object that defines parameters to control the formatting and redirection of received SMTP
    messages. Whenever an SMTP message is received, Mappings are checked by the `Listener` to see if the received SMTP
    message headers match the regex defined in the Mapping object.
    """
    # Private attributes are not for public consumption.
    # pylint: disable=too-many-instance-attributes

    # Initialize attributes
    _pattern = None
    _field = None
    _connector = None
    _parser = None

    def __init__(self, pattern: str, connector, **kwargs):
        """
        Initializes the Mapping object with desired attributes.

        Args:
            pattern (str): The regex pattern to use when checking SMTP headers for specific values.
            connector (mail2beyond.framework.BaseConnector): The Connector object this Mapping will use whenever a
                match is found. This must be an object of a class that has a base class of
                `mail2beyond.framework.BaseConnector`. You should pass in a built-in `Connector` object created from the
                mail2beyond.connectors module or your own custom `Connector` object.
            **field (str) The SMTP header to run the regex pattern against. This is commonly the TO or FROM headers to
                check the email's recipient or sender respectively, but can be any header available in the received
                email. Defaults to `from`.
            **parser (mail2beyond.framework.BaseParser): The Parser class to use whenever this Mapping is matched. This
                allows you to specify a Parser class that will parse the email's content body to a more human-readable
                format. Note this must be the Parser class NOT a `Parser` object. This must be Parser class with a base
                class of `mail2beyond.framework.BaseParser`. Defaults to the `mail2beyond.framework.BaseParser` class,
                but it is strongly recommended you pass in `mail2beyond.parsers.auto.Parser` to automatically parse
                the content body based on the email's content-type header or pass in your own custom parser class.

        See Also:
            https://github.com/soluna-studios/mail2beyond/blob/master/docs/PACKAGE.md#writing-custom-connectors
            https://github.com/soluna-studios/mail2beyond/blob/master/docs/PACKAGE.md#writing-custom-parsers
        """
        # Assign required attributes
        self.pattern = pattern
        self.connector = connector

        # Assign optional attributes, or assume defaults.
        self.field = kwargs.get("field", "from")
        self.parser = kwargs.get("parser", BaseParser)

    def __str__(self):
        """Sets the object's pattern as the string representation of this object."""
        return self.pattern

    def is_match(self, value):
        """
        Checks if a specified value matches this mapping.

        Args:
            value (str): The value to check for a regex pattern match.

        Returns:
            bool: Whether the value was a match for the regex pattern or not.
        """
        # Check if the value matches this mapping's regex pattern.
        if value is not None and re.search(self.pattern, value):
            return True
        # No match, return False
        return False

    # Getters and setters #
    @property
    def pattern(self):
        """The property to get and/or set the  pattern attribute."""
        return self._pattern

    @pattern.setter
    def pattern(self, value: str):
        """Sets the pattern attribute after validating the new value."""
        # Require pattern to be a string
        if isinstance(value, str):
            self._pattern = value
        else:
            raise TypeError("pattern must be type 'str'")

    @property
    def connector(self):
        """The property to get and/or set the  connector attribute."""
        return self._connector

    @connector.setter
    def connector(self, value):
        """Sets the connector attribute after validating the new value."""
        # Require connector to have a base class of mail2beyond.framework.BaseConnector
        if value.__class__.__base__ == BaseConnector:
            self._connector = value
        else:
            raise TypeError("connector must be object with base class 'BaseConnector'")

    @property
    def parser(self):
        """The property to get and/or set the  parser attribute."""
        return self._parser

    @parser.setter
    def parser(self, value):
        """Sets the parser attribute after validating the new value."""
        # Require parser to have a base class of mail2beyond.framework.BaseParser
        if inspect.isclass(value) and issubclass(value, BaseParser):
            self._parser = value
        else:
            raise TypeError("parser must be a class (not object) with base class 'BaseParser'")

    @property
    def field(self):
        """The property to get and/or set the  field attribute."""
        return self._field

    @field.setter
    def field(self, value: str):
        """Sets the field attribute after validating the new value."""
        # Require field to be a support/recognized field
        if isinstance(value, str):
            self._field = value
        else:
            raise ValueError("field must be type 'str'")


class BaseConnector:
    """
    Creates the BaseConnector object to be extended by configurable child classes. This establishes standard
    methods and attributes for creating plugin API connectors.

    Attributes:
        name (str): The name of the created object. This attribute is primarily used for the CLI, but could be useful
            for other reasons.
        log (logging.Logger): The Logger object to use when logging events that occur when the `Connector` is called.
        config (dict): The dict of custom `Connector` configuration values. Any arguments passed in when the object
            is initially created will be populated here. Some Connectors will require configurable values like URLs,
            credentials, etc. This is where those values should be passed in.
    """
    # Attributes
    name = ""

    def __init__(self, **kwargs):
        """
        Initialize the object with required attributes.

        Notes:
            Any arguments passed in when this object is created will be stored in the 'config' attribute of the object.
        """
        self.config = kwargs
        self.log = logging.getLogger(__name__)

    def __str__(self):
        """Sets this object's name attribute as its string representation."""
        return self.name

    def run(self, parser):
        """
        Runs the current connector object. This method calls the `pre_submit()` and `submit()` methods respectively and
        checks for any errors encountered.

        Args:
            parser (mail2beyond.framework.BaseParser): The `Parser` object to use when this `Connector` is called. This
                will the object created by the mapping's parser and will contain the parsed email content.
                Note this must be the `Parser` object NOT a `Parser` class. This must be `Parser` object with a base
                class of `mail2beyond.framework.BaseParser`.

        Raises:
            mail2beyond.framework.Error: When the `pre_submit()` method catches a `mail2beyond.framework.Error`.
            Exception: When the `submit()` method catches any error, the error will simply be re-raised by this method.
        """
        # Try to run pre-submit checks and log errors.
        try:
            self.pre_submit(parser)
        except Error as pre_submit_err:
            self.log.error(f"pre-submit checks for connector '{self}' failed ({pre_submit_err})")
            raise pre_submit_err

        # Try to run a submit and log errors
        try:
            self.submit(parser)
        except Exception as submit_error:
            self.log.error(f"submit for connector '{self}' failed ({submit_error})")
            raise submit_error

    def submit(self, parser):
        """
        Initializes the `submit()` method that performs actions required for the connector redirect the SMTP message.
        This method is intended to be overwritten by a child class that creates a `Connector` object for a specific API
        or service. If this method is not overwritten by the child class, an error is raised. When overwriting this
        method, you can use the `parser` object to reference various components in the received SMTP message and the
        `config` attribute to reference required configuration values. You can also log events in this method using
        the `log` attribute.

        Args:
            parser (mail2beyond.framework.BaseParser): The `Parser` object to use when this `Connector` is called. This
                will the object created by the mapping's parser and will contain the parsed email content.
                Note this must be the `Parser` object NOT a `Parser` class. This must be `Parser` object with a base
                class of `mail2beyond.framework.BaseParser`.

        Raises:
            mail2beyond.framework.Error: When the `submit()` method has not been overwritten by a child class.
        """
        raise Error(f"method has not been overwritten by child class but received {parser}")

    def pre_submit(self, parser):
        """
        Initializes the `pre_submit()` method that performs validation before the connector's `submit()` method is
        called.This method is intended to be overwritten by a child class that creates `Connector` objects for a
        specific API. Typically, ths method will be used to validate the contents of the `Connector` object's `config`
        attribute. But could also be used to validate the parsed email values using the `parser` attribute. When
        overwriting this method, simply raise a `mail2beyond.framework.Error` to mark the `pre_submit()` checks as a
        failure. You can also log events in this method using the `log` attribute.

        Args:
            parser (mail2beyond.framework.BaseParser): The `Parser` object to use when this `Connector` is called. This
                will the object created by the mapping's parser and will contain the parsed email content.
                Note this must be the `Parser` object NOT a `Parser` class. This must be `Parser` object with a base
                class of `mail2beyond.framework.BaseParser`.

        Returns:
            mail2beyond.framework.BaseParser: Simply returns the `Parser` object by default,
        """
        return parser


class BaseParser:
    """
    Creates a `BaseParser` object that can be used to further parse an `Email` object's content body.

    Attributes:
        name (str): The name of the object. This can be used to differentiate different parser objects from each other.
        mail (mail2beyond.framework.Email): The `Email` object containing the decoded email contents.
        log (logging.Logger): The Logger object to use when logging events that occur when the Parser is called.
        config (dict): The dict of custom `Parser` configuration values. Any extra arguments passed in when the object
            is initially created will be populated here. This can be used to define configurable attributes to your
            custom Parser classes.
    """
    name = ""
    _mail = None
    _config = None

    def __init__(self, mail: Email, **kwargs):
        """
        Initializes the `Parser` object with required attributes.

        Args:
            mail (mail2beyond.framework.Email): The `Email` object containing the decoded email contents.

        Notes:
            Any extra arguments passed in when this object is created will be stored in the 'config' attribute of the
                object.
        """
        self.mail = mail
        self.config = kwargs
        self.log = logging.getLogger(__name__)

    def __str__(self):
        """Sets the string representation of this object."""
        return self.name

    def parse_content(self):
        """
        Initializes the `Parser` object's `parse_content()` method. This method is intended to be overwritten by a child
        class to add parsers for various formats.

        Returns:
            str: By default, this method will simply return the current content decoded content from the 'mail'
                attribute. This method is intended to be overwritten by a child class to extend functionality.
        """
        return self.mail.content

    # Getters and setters
    @property
    def mail(self):
        """The property to get and/or set the  mail attribute."""
        return self._mail

    @mail.setter
    def mail(self, value: Email):
        """Sets the mail attribute after validating the new value."""
        # Ensure value is a mail2beyond.framework.Email object
        if not isinstance(value, Email):
            raise Error("'parser' must be type 'mail2beyond.framework.Email'")

        self._mail = value

    @property
    def subject(self):
        """The property to get and/or set the  subject attribute."""
        return self.mail.headers.get("subject", "No subject")

    @property
    def content(self):
        """Gets the parsed content. This is essentially a shortcut for calling parse_content()."""
        return self.parse_content()

    @property
    def config(self):
        """The property to get and/or set the  config attribute."""
        return self._config

    @config.setter
    def config(self, value: dict):
        """Sets the config attribute after validating the new value."""
        # Ensure config is a dict
        if not isinstance(value, dict):
            raise Error("'config' must be type 'dict'")

        self._config = value



"""Creates the built-in 'void' connector that can be used to discard SMTP messages."""
from mail2beyond import framework


class Connector(framework.BaseConnector):
    """Defines a connector that does not do anything."""
    name = "void"

    def submit(self, parser):
        """Overwrites the submit() method, but does nothing."""
        self.log.debug(f"connector '{self}' successfully sent message to the abyss")

    def pre_submit(self, parser):
        """Overwrites the pre_submit() method, but does nothing."""



"""
Module that contains a parser that parse an email's content body as plaintext. This simply extends the
mail2beyond.framework.BaseParser class but retains its default functionality.
"""

from mail2beyond import framework


class Parser(framework.BaseParser):
    """Creates a parser to parse an email's content body as plaintext."""
    name = "text"



"""Module used to setup and install the mail2beyond package."""

import os
import codecs

from setuptools import setup


def read(rel_path):
    """Reads a specified file."""
    here = os.path.abspath(os.path.dirname(__file__))
    with codecs.open(os.path.join(here, rel_path), 'r') as filepath:
        return filepath.read()


def get_readme():
    """Opens the README.md file for this package so it can by used in setup.py."""
    # Read the readme file
    return read("README.md")


def get_requirements():
    """Opens the requirements.txt file for this package so it can be used in setup.py."""
    reqs = read("requirements.txt").split("\n")

    # Remove empty items if any
    while "" in reqs:
        reqs.remove("")

    return reqs


def get_version(rel_path):
    """
    Gets the current version of the package. If a __MAIL2BEYOND_REVISION__ environment variable exists, it will
    be read and appended to the current package version. This is used to ensure the setup version can always be unique
    for PyPI dev builds triggered by CI/CD workflows.
    """
    # Variables
    revision = ""

    # If a __MAIL2BEYOND_REVISION__ environment variable exists, set it as the dev revision.
    if "__MAIL2BEYOND_REVISION__" in os.environ:
        revision = "." + os.environ.get("__MAIL2BEYOND_REVISION__")

    # Otherwise, look for the version in the package.
    for line in read(rel_path).splitlines():
        if line.startswith('__version__'):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1] + revision

    raise RuntimeError("Unable to find version string.")


setup(
    name='mail2beyond',
    author='Jared Hendrickson',
    author_email='jaredhendrickson13@gmail.com',
    url="https://github.com/soluna-studios/mail2beyond",
    license="MIT",
    description="A Python based SMTP server package and CLI that redirects incoming SMTP messages to upstream APIs like"
                " Google Chat, Slack and more!.",
    long_description=get_readme(),
    long_description_content_type="text/markdown",
    version=get_version("mail2beyond/__init__.py"),
    scripts=['scripts/mail2beyond'],
    packages=["mail2beyond", "mail2beyond.connectors", "mail2beyond.parsers"],
    install_requires=get_requirements(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.8'
)



"""
Creates the built-in 'slack' connector that can be used to translate SMTP messages into Slack messages.
"""

import requests

from mail2beyond import framework


class Connector(framework.BaseConnector):
    """Defines a connector that allows mail2beyond to integrate with Slack."""
    name = "slack"

    def submit(self, parser):
        """Overwrites the submit() method to send a Slack message."""
        # Submit this message to Slack via webhook URL
        try:
            resp = requests.post(
                url=self.config["webhook_url"],
                headers={'Content-Type': 'application/json'},
                json={"text": f"*{parser.subject}*\n\n{parser.content}"}
            )
            self.log.debug(f"connector '{self}' responded with {resp.status_code} status {resp.text}")
        except Exception as slack_req_err:
            self.log.error(f"connector '{self}' failed '{slack_req_err}'")
            raise framework.Error(f"connector '{self}' : {slack_req_err}")

    def pre_submit(self, parser):
        """Overwrites the pre_submit() method to ensure required config values are set."""
        # Ensure ths connector has a config value for 'webhook_url'
        if "webhook_url" not in self.config:
            raise framework.Error(f"connector '{self}' requires config value 'webhook_url'")
        # Ensure the webhook_url config value is a string
        if not isinstance(self.config["webhook_url"], str):
            raise framework.Error(f"connector '{self}' config value 'webhook_url' must be type 'str'")



"""
Module that includes tool functions that primarily assist the CLI and unit tests, but may be useful to others.
"""

import inspect
import logging
import pathlib
import ssl
import sys

from OpenSSL import crypto

import mail2beyond.framework
from . import connectors
from . import parsers
from . import framework


def get_connector_modules(path: (str, None) = None):
    """
    Gathers all available connector modules. This allows a 'path' to be specified to optionally pass in plugin connector
    modules. Built-in connectors are always included.

    Args:
        path (str, None): A path to a directory that contains plugin connector modules. Only .py files within this
            directory will be included. Each .py file must include a class named `Connector` that extends the
            `mail2beyond.framework.BaseConnector class. If `None` is specified, only the built-in connector modules
            will be available.

    Raises:
        mail2beyond.framework.Error: When plugin connector modules could not be loaded.

    Returns:
        dict: A dictionary of available connector modules. The dictionary keys will be the module names and the values
            will be the module itself.
    """
    # Start by gathering the built-in connectors from the mail2beyond.connectors sub-package.
    available_connectors = dict(inspect.getmembers(connectors, inspect.ismodule))

    # If a plugin path was passed in, include modules within that directory as well.
    if path:
        # Convert 'path' into an object
        path_obj = pathlib.Path(path)

        # Require path to be an existing directory
        if not path_obj.exists() or not path_obj.is_dir():
            raise framework.Error(f"failed to load connector modules '{path}' is not an existing directory")

        # Add this directory to our Python path
        sys.path.append(str(path_obj.absolute()))

        # Loop through each .py file in the directory and ensure it is valid
        for module_path in path_obj.glob("*.py"):
            # Verify this module could be imported, contains the Connector class and is added to available connectors.
            try:
                module = __import__(module_path.stem)
                getattr(module, "Connector")
            except ModuleNotFoundError as exc:
                mod_not_found_err_msg = f"failed to import connector module '{module_path.stem}' from '{path}'"
                raise framework.Error(mod_not_found_err_msg) from exc
            except AttributeError as exc:
                attr_err_msg = f"connector module '{module_path.stem}' from '{path}' has no class named 'Connector'"
                raise framework.Error(attr_err_msg) from exc

            # Ensure the module's Connector class is a subclass of BaseConnector
            if inspect.isclass(module.Connector) and issubclass(module.Connector, mail2beyond.framework.BaseConnector):
                available_connectors[module_path.stem] = module
                continue

            # Throw an error if the module's Connector class is not a subclass of BaseConnector
            raise framework.Error(
                f"'Connector' class in '{ module_path }' is not subclass of 'mail2beyond.framework.BaseConnector'"
            )

    # Return the gathered connector modules
    return available_connectors


def get_connectors_from_dict(config: dict, path: (str, None) = None):
    """
    Converts a dictionary representations of connectors to connector objects.
    Args:
        config (dict): A dictionary representation of different connectors to create.
        path (str, None): A path to a directory that contains plugin connector modules. Only .py files within this
            directory will be included. Each .py file must include a class named `Connector` that extends the
            `mail2beyond.framework.BaseConnector` class. If `None` is specified, only the built-in connector modules
            will be available.

    Raises:
        mail2beyond.framework.Error: When a validation error occurs.

    Returns:
        list: A list of Connector objects that can be used.
    """
    # Create a list to store the created connector objects
    valid_connectors = []
    available_connectors = get_connector_modules(path)

    # Require connectors config to be defined
    if "connectors" not in config.keys():
        raise framework.Error("'connectors' value is required")

    # Require connectors config to be a list
    if not isinstance(config.get("connectors"), list):
        raise framework.Error("'connectors' value must be type list")

    # Require at least one connector
    if len(config.get("connectors")) == 0:
        raise framework.Error("at least one 'connectors' item is required")

    # Loop through each connector in the configuration
    for connector in config.get("connectors"):
        # Require connector definition to be a dict
        if not isinstance(connector, dict):
            raise framework.Error("'connectors' items must be type dict")

        # Require connector name to be defined
        if "name" not in connector.keys():
            raise framework.Error("'connectors' items must contain 'name' value")

        # Require connector names to be unique
        for obj in valid_connectors:
            if connector["name"] == obj.name:
                raise framework.Error(f"multiple 'connectors' items assigned name '{obj.name}'")

        # Require connector type to be defined
        if "module" not in connector.keys():
            raise framework.Error("'connectors' configuration items must contain 'module' value")

        # Require connector type to be known
        if connector["module"] not in available_connectors:
            raise framework.Error("'connectors' configuration items must contain 'module' value")

        # Get the connector module that matches this connector
        connector_module = available_connectors.get(connector["module"])

        # Add this connector, it is valid
        connector_obj = connector_module.Connector()
        connector_obj.name = connector.get("name")
        connector_obj.config = connector.get("config", {})

        valid_connectors.append(connector_obj)

    # Return our connector objects
    return valid_connectors


def get_connector_by_name(name, connector_objs):
    """
    Gets the connector with a specific name from a list of Connector objects.

    Args:
        name (str): The name of the connector to find.
        connector_objs (list): A list of Connector objects to search.

    Returns:
        mail2beyond.framework.BaseConnector: The Connector object with the specified name if one was found.
        None: When no Connector object was found with the specified name.
    """
    # Loop through each connector and return the connector with the specified name
    for connector in connector_objs:
        if connector.name == name:
            return connector

    return None


def get_parser_modules(path: (str, None) = None):
    """
    Gathers all available parser modules. This allows a 'path' to be specified to optionally pass in plugin parser
    modules. Built-in parsers are always included.

    Args:
        path (str, None): A path to a directory that contains plugin parser modules. Only .py files within this
            directory will be included. Each .py file must include a class named `Parser` that extends the
            `mail2beyond.framework.BaseParser` class. If `None` is specified, only the built-in parser modules
            will be available.

    Raises:
        mail2beyond.framework.Error: When plugin parser modules could not be loaded.

    Returns:
        dict: A dictionary of available parser modules. The dictionary keys will be the module names and the values
            will be the module itself.
    """
    # Start by gathering the built-in parsers from the mail2beyond.connectors sub-package.
    available_parsers = dict(inspect.getmembers(parsers, inspect.ismodule))

    # If a plugin path was passed in, include modules within that directory as well.
    if path:
        # Convert 'path' into an object
        path_obj = pathlib.Path(path)

        # Require path to be an existing directory
        if not path_obj.exists() or not path_obj.is_dir():
            raise framework.Error(f"failed to load parser modules '{path}' is not an existing directory")

        # Add this directory to our Python path
        sys.path.append(str(path_obj.absolute()))

        # Loop through each .py file in the directory and ensure it is valid
        for module_path in path_obj.glob("*.py"):
            # Verify this module could be imported, contains the Parser class and is added to available connectors.
            try:
                module = __import__(module_path.stem)
                getattr(module, "Parser")
                available_parsers[module_path.stem] = module
            except ModuleNotFoundError as exc:
                mod_not_found_err_msg = f"failed to import parser module '{module_path.stem}' from '{path}'"
                raise framework.Error(mod_not_found_err_msg) from exc
            except AttributeError as exc:
                attr_err_msg = f"parser module '{module_path.stem}' from '{path}' has no class named 'Parser'"
                raise framework.Error(attr_err_msg) from exc

            # Ensure the module's Parser class is a subclass of BaseParser
            if inspect.isclass(module.Parser) and issubclass(module.Parser, mail2beyond.framework.BaseParser):
                available_parsers[module_path.stem] = module
                continue

            # Throw an error if the module's Parser class is not a subclass of BaseParser
            raise framework.Error(
                f"'Parser' class in '{module_path}' is not subclass of 'mail2beyond.framework.BaseParser'"
            )

    # Return the gathered parser modules
    return available_parsers


def get_mappings_from_dict(config: dict, connectors_path: (str, None) = None, parsers_path: (str, None) = None):
    """
    Converts a dictionary representations of mappings to Mapping objects. Since Mapping objects are dependent on
    a Connector object, the dictionary must also include representations of Connector objects to use.

    Args:
        config (dict): A dictionary representation of different mappings to create.
        connectors_path (str, None): A path to a directory that contains plugin connector modules. Only .py files within
            this directory will be included. Each .py file must include a class named `Connector` that extends the
            `mail2beyond.framework.BaseConnector` class. If `None` is specified, only the built-in connector modules
            will be available.
        parsers_path (str, None): A path to a directory that contains plugin parser modules. Only .py files within this
            directory will be included. Each .py file must include a class named `Parser` that extends the
            `mail2beyond.framework.BaseParser` class. If `None` is specified, only the built-in parser modules
            will be available.

    Raises:
        mail2beyond.framework.Error: When a validation error occurs.

    Returns:
        list: A list of Mapping objects that can be used.
    """
    # Variables
    config_connectors = get_connectors_from_dict(config, path=connectors_path)
    available_parsers = get_parser_modules(path=parsers_path)
    valid_mappings = []

    # Require mappings config to be defined
    if "mappings" not in config.keys():
        raise framework.Error("'mappings' value is required")

    # Require mappings config to be a list
    if not isinstance(config.get("mappings"), list):
        raise framework.Error("'mappings' value must be type list")

    # Require at least one mapping
    if len(config.get("mappings")) == 0:
        raise framework.Error("at least one 'mappings' item is required")

    # Loop through each mapping in the configuration
    for mapping in config.get("mappings"):
        # Require mapping definition to be a dict
        if not isinstance(mapping, dict):
            raise framework.Error("'mappings' items must be type dict")

        # Require mapping connector to be defined
        if "connector" not in mapping.keys():
            raise framework.Error("'mappings' items must contain 'connector' value")

        # Get this mapping's connector
        connector = get_connector_by_name(mapping.get("connector"), config_connectors)

        # Require this mapping's connector to exist
        if not connector:
            raise framework.Error(f"'mappings' item references undefined connector '{mapping['connector']}'")

        # Replace the mapping's connector item with the actual object so it can be passed in with kwargs below
        mapping["connector"] = connector

        # Ensure parser is a known
        parser = mapping.get("parser", "auto")
        if parser not in available_parsers:
            raise framework.Error(f"'mappings' item references undefined parser module '{parser}'")

        # Replace the mapping's parser item with the actual Parser class so it can be passed into with kwargs below
        mapping["parser"] = available_parsers.get(parser).Parser

        # Add this mapping, it is valid
        mapping_obj = framework.Mapping(**mapping)
        valid_mappings.append(mapping_obj)

    # Return our connector objects
    return valid_mappings


def get_listeners_from_dict(config: dict, log_level: int = logging.NOTSET, **kwargs):
    """
    Converts a dictionary representations of listeners to Listener objects. Since Listener objects are dependent on
    a Mapping objects, and Mapping objects are dependent on Connector objects, the dictionary must also include
    representations of both Mapping and Connector objects to use.

    Args:
        config (dict): A dictionary representation of different mappings to create.
        log_level (int): Sets the logging level the Listeners' Logger will start logging at. See
            https://docs.python.org/3/library/logging.html#logging-levels
        **connectors_path (str, None): A path to a directory that contains plugin connector modules. Only .py files
            within this directory will be included. Each .py file must include a class named `Connector` that extends
            the`mail2beyond.framework.BaseConnector` class. If `None` is specified, only the built-in connector modules
            will be available.
        **parsers_path (str, None): A path to a directory that contains plugin parser modules. Only .py files within
            this directory will be included. Each .py file must include a class named `Parser` that extends the
            `mail2beyond.framework.BaseParser` class. If `None` is specified, only the built-in parser modules
            will be available.

    Raises:
        mail2beyond.framework.Error: When a validation error occurs.

    Returns:
        list: A list of Listener objects that can be used.
    """
    # Validating listeners from configuration requires many conditions because there are many options.
    # In the future, consider grouping config validation into it's own class with getters and setters.
    # pylint: disable=too-many-branches,too-many-statements

    # Constants
    tls_protocols = {
        "tls1": ssl.TLSVersion.TLSv1,
        "tls1_1": ssl.TLSVersion.TLSv1_1,
        "tls1_2": ssl.TLSVersion.TLSv1_2,
        "tls1_3": ssl.TLSVersion.TLSv1_3
    }
    enable_smtps_default = False
    enable_starttls_default = False
    require_starttls_default = False
    tls_minimum_version_options = tls_protocols.keys()
    tls_minimum_version_default = "tls1_2"

    # Variables
    valid_listeners = []
    mappings = get_mappings_from_dict(
        config,
        connectors_path=kwargs.get("connectors_path"),
        parsers_path=kwargs.get("parsers_path")
    )

    # Require listeners config to be defined.
    if "listeners" not in config.keys():
        raise framework.Error("'listeners' value is required")

    # Require listeners config to be a list
    if not isinstance(config.get("listeners"), list):
        raise framework.Error("'listeners' value must be type list")

    # Require at least one listener
    if len(config.get("listeners")) == 0:
        raise framework.Error("at least one 'listeners' item is required")

    # Loop through each listener and ensure it has a valid configuration.
    for listener in config.get("listeners"):
        # Reset any previous SSLContext
        context = None
        tls_listener = False

        # Require listener definition to be a dict
        if not isinstance(listener, dict):
            raise framework.Error("'listeners' items must be type dict")

        # Ensure listener has an address and port set. The Listener object will validate further upon creation.
        if "address" not in listener.keys():
            raise framework.Error("'listeners' items must contain 'address' value")

        # Ensure listener has an address and port set. The Listener object will validate further upon creation.
        if "port" not in listener.keys():
            raise framework.Error("'listeners' items must contain 'port' value")

        # When 'enable_smtps' is set, ensure it is a bool
        if not isinstance(listener.get("enable_smtps", enable_smtps_default), bool):
            raise framework.Error("'enable_smtps' items must be type bool")

        # Mark this as a TLS enabled listen if SMTPS is enabled
        if listener.get("enable_smtps", enable_smtps_default):
            tls_listener = True

        # When 'enable_starttls' is set, validate associated fields
        if "enable_starttls" in listener.keys():
            # Mark this as a TLS enabled listener
            tls_listener = True

            # Ensure 'enable_smtps' was not also enabled
            if listener.get("enable_smtps", enable_smtps_default):
                raise framework.Error(
                    "'listeners' items 'enable_starttls' and 'enable_smtps' cannot be active on the same listener"
                )

            # Ensure it is a bool
            if not isinstance(listener.get("enable_starttls", enable_starttls_default), bool):
                raise framework.Error("'listeners' item 'enable_starttls' value must be type bool")

            # Check for 'require_starttls' option and ensure it is a bool
            if not isinstance(listener.get("require_starttls", require_starttls_default), bool):
                raise framework.Error("'listeners' item 'require_starttls' value must be type bool")

        # If this was marked as a TLS listener, validate TLS configuration
        if tls_listener:
            # Require the minimum_tls_protocol to be known
            if listener.get("minimum_tls_protocol", tls_minimum_version_default) not in tls_minimum_version_options:
                raise framework.Error("'listeners' item 'minimum_tls_protocol' value is not known protocol")

            # Require a TLS certificate and key
            for field in ["tls_cert", "tls_key"]:
                if field not in listener.keys():
                    raise framework.Error(f"'listeners' item '{field}' is required to use stmps or starttls")

                # Ensure field is string
                if not isinstance(listener.get(field), str):
                    raise framework.Error(f"'listeners' item '{field}' value must be type str")

                # Ensure field is an existing filepath
                if not pathlib.Path(listener.get(field)).exists():
                    raise framework.Error(f"'listeners' item '{field}' is not existing file path")

            # Create the SSLContext for this listener
            context = ssl.SSLContext(ssl.PROTOCOL_TLS)
            context.load_cert_chain(listener.get("tls_cert"), listener.get("tls_key"))
            context.set_ciphers("HIGH")
            context.minimum_version = tls_protocols[listener.get("tls_minimum_version", tls_minimum_version_default)]

        # Create an SMTPS listener if configured
        if listener.get("enable_smtps", enable_smtps_default):
            valid_listeners.append(
                framework.Listener(
                    mappings=mappings,
                    address=listener.get("address"),
                    port=listener.get("port"),
                    tls_context=context,
                    enable_starttls=False,
                    require_starttls=False,
                    log_level=log_level
                )
            )
        # Create a STARTTLS listener if configured
        elif listener.get("enable_starttls", enable_starttls_default):
            valid_listeners.append(
                framework.Listener(
                    mappings=mappings,
                    address=listener.get("address"),
                    port=listener.get("port"),
                    tls_context=context,
                    enable_starttls=True,
                    require_starttls=listener.get("require_starttls", require_starttls_default),
                    log_level=log_level
                )
            )
        # Create normal SMTP listener otherwise
        else:
            valid_listeners.append(
                framework.Listener(
                    mappings=mappings,
                    address=listener.get("address"),
                    port=listener.get("port"),
                    log_level=log_level
                )
            )

    # Return the validated listeners
    return valid_listeners


def generate_tls_certificate(cert_path: str, key_path: str):
    """
    Generates a self-signed certificate and private key at a specified file path. This is primarily used for unit tests,
    but could be useful elsewhere too.
    Args:
        cert_path (str): The file path to write the generated certificate file to.
        key_path (str): The file path to write the generated key file to.
    """

    # Generate private key
    key = crypto.PKey()
    key.generate_key(crypto.TYPE_RSA, 2048)

    # Define self-signed certificate
    cert = crypto.X509()
    cert.get_subject().CN = "example.com"
    cert.set_serial_number(0)
    cert.gmtime_adj_notBefore(0)
    cert.gmtime_adj_notAfter(365000)
    cert.set_issuer(cert.get_subject())
    cert.set_pubkey(key)
    cert.sign(key, 'sha512')

    # Write the certificate and key to file
    with open(cert_path, "wt", encoding="utf-8") as cert_file:
        cert_file.write(crypto.dump_certificate(crypto.FILETYPE_PEM, cert).decode("utf-8"))
    with open(key_path, "wt", encoding="utf-8") as key_file:
        key_file.write(crypto.dump_privatekey(crypto.FILETYPE_PEM, key).decode("utf-8"))
